@@ -1,24 +1,50 @@
-"""Basic classes for creating a pygame application"""
+'''Basic classes for creating a pygame application'''
 
 import pygame, math, sys
 from string import printable as _printable
-from typing import List
+from typing import List, Type, TypeVar, Tuple
 from glob import glob
 from pygame.locals import *
 from recordclass import RecordClass
 
 printable = _printable.strip() + ' '
 
+T = TypeVar('T', bound='ManyOf')
+
+class ManyOf:
+    '''
+    A class that can be used to group objects and call them at the same time
+    reads methods from a given class and creates those methods in this class
+    by looping over each provided object and calling that method on it
+    '''
+    def __init__(self, cls: Type[T], *obj_list: Tuple[T]):
+        '''
+        initialize the ManyOf class
+        :cls: the type or class of the objects in obj_list
+        :obj_list: the list of objects that the methods will be called on
+        '''
+        if len(obj_list) < 1:
+            raise ValueError('Must pass at least one object in obj_list')
+        self.obj_list = obj_list
+        for method in filter(lambda x: x != '__class__', dir(cls)):
+            if not callable(getattr(cls, method)): # if the method is not a function
+                continue # move on to the next method
+            def func(*args, __method_name__=method, **kwargs): # create new function
+                results = [getattr(obj, __method_name__)(*args, **kwargs) for obj in obj_list]
+                return ManyOf(type(results[0]), *results)
+            func.__name__ = method # rename function
+            setattr(self, method, func) # save function as method
+
 class TrueEvery:
-    """This is a functor that creates a function that returns true once every {self.count} calls"""
+    '''This is a functor that creates a function that returns true once every {self.count} calls'''
 
     def __init__(self, count: int, initial_count: int = None, once: bool = False, start_value: int = 0):
-        """
+        '''
         :count: the number of times {self.__call__} must be called to return true once
         :initial_count: Optional. defaults to {self.count}. the number of times {self._call__} must be called to return True after the first call
         :once: Optional. defaults to False. the value
         :start_value: Optional. defaults to 0. the value that the offset starts at before the current call
-        """
+        '''
         self.count = count
         self.initial_count = initial_count if initial_count != None else count
         self.once = once
@@ -26,11 +52,11 @@ class TrueEvery:
         self.first_call = True
 
     def __call__(self) -> bool:
-        """
+        '''
         Override () operator
         :returns: true once every {self.count} calls
             always returns true first time run unless start_value is set to something different
-        """
+        '''
         # TODO: refactor this
         if not self.first_call and self.once:
             return False
@@ -41,19 +67,19 @@ class TrueEvery:
             return True
 
     def reset(self, override_start_value: int = None):
-        """
+        '''
         reset {self.calls}, and {self.first_call}
         :override_start_value: Optional. Defaults to self.start_value. set a new start value instead of the one in the constructor
-        """
+        '''
         self.calls = override_start_value if override_start_value != None else self.start_value
         self.first_call = True
 
     def run_or_reset(self, boolean: bool) -> bool:
-        """
+        '''
         :boolean: the boolean to be evaluated. If this boolean is True the {self.__call__} is called.
             If the boolean is False it calls {self.reset}
         :returns: a bool. It returns the result of call if {boolean} is True. or it returns False if {boolean} is False
-        """
+        '''
         if boolean:
             return self()
         self.reset()
@@ -72,68 +98,88 @@ class Point(RecordClass):
         '''
         return Point(-self.x, -self.y)
 
-    def __add__(self, pos: Point) -> Point:
+    def __add__(self, other: int | float | Point) -> Point:
         '''add two points'''
-        if not isinstance(pos, Point):
-            pos = Point._make(pos)
-        return Point(self.x + pos.x, self.y + pos.y)
+        if isinstance(other, int | float):
+            return Point(self.x + other, self.y + other)
+        if not isinstance(other, Point):
+            try:
+                other = Point._make(other)
+            except TypeError:
+                return NotImplemented
+        return Point(self.x + other.x, self.y + other.y)
 
-    def __radd__(self, pos: Point) -> Point:
+    def __radd__(self, other: int | float | Point) -> Point:
         '''add two points'''
-        return self + pos
+        return self + other
 
-    def __sub__(self, pos: Point) -> Point:
-        '''subtract self from pos'''
-        if not isinstance(pos, Point):
-            pos = Point._make(pos)
-        return self + (-pos)
+    def __sub__(self, other: int | float | Point) -> Point:
+        '''subtract self from other'''
+        if not isinstance(other, int | float) and not isinstance(other, Point):
+            other = Point._make(other)
+        return self + (-other)
 
-    def __rsub__(self, pos: Point) -> Point:
-        '''subtract self from pos'''
-        return -self + pos
+    def __rsub__(self, other: int | float | Point) -> Point:
+        '''subtract self from other'''
+        return -self + other
 
-    def __mul__(self, other: float | int | Point) -> Point:
+    def __mul__(self, other: int | float | Point) -> Point:
         '''Multiply x and y by other'''
-        if isinstance(other, float| int):
+        if isinstance(other, int | float):
             return Point(self.x * other, self.y * other)
         if not isinstance(other, Point):
-            other = Point._make(other)
+            try:
+                other = Point._make(other)
+            except TypeError:
+                return NotImplemented
         return Point(self.x * other.x, self.y * other.y)
 
-    def __rmul__(self, other: float | int | Point) -> Point:
+    def __rmul__(self, other: int | float | Point) -> Point:
         '''Multiply x and y by other'''
         return self * other
 
-    def __truediv__(self, other: float | int | Point) -> Point:
+    def __truediv__(self, other: int | float | Point) -> Point:
         '''Divide x and y by other'''
-        if isinstance(other, float | int):
+        if isinstance(other, int | float):
             return Point(self.x / other, self.y / other)
         if not isinstance(other, Point):
-            other = Point._make(other)
+            try:
+                other = Point._make(other)
+            except TypeError:
+                return NotImplemented
         return Point(self.x / other.x, self.y / other.y)
 
-    def __rtruediv__(self, other: float | int | Point) -> Point:
+    def __rtruediv__(self, other: int | float | Point) -> Point:
         '''Divide other by x and y'''
-        if isinstance(other, float | int):
-            return Point(other / self.x, other / self.y)
+        if isinstance(other, int | float):
+            return Point(other / self.x , other / self.y)
         if not isinstance(other, Point):
-            other = Point._make(other)
+            try:
+                other = Point._make(other)
+            except TypeError:
+                return NotImplemented
         return Point(other.x / self.x, other.y / self.y)
 
-    def __floordiv__(self, other: float | int | Point) -> Point:
+    def __floordiv__(self, other: int | float | Point) -> Point:
         '''Divide x and y by other and round down'''
-        if isinstance(other, float | int):
+        if isinstance(other, int | float):
             return Point(self.x // other, self.y // other)
         if not isinstance(other, Point):
-            other = Point._make(other)
+            try:
+                other = Point._make(other)
+            except TypeError:
+                return NotImplemented
         return Point(self.x // other.x, self.y // other.y)
 
-    def __rfloordiv__(self, other: float | int | Point) -> Point:
+    def __rfloordiv__(self, other: int | float | Point) -> Point:
         '''Divide other by x and y and round down'''
-        if isinstance(other, float | int):
-            return Point(other // self.x, other // self.y)
+        if isinstance(other, int | float):
+            return Point(other // self.x , other // self.y)
         if not isinstance(other, Point):
-            other = Point._make(other)
+            try:
+                other = Point._make(other)
+            except TypeError:
+                return NotImplemented
         return Point(other.x // self.x, other.y // self.y)
 
     def __floor__(self) -> Point:
@@ -147,12 +193,19 @@ class Point(RecordClass):
     def __eq__(self, pos: Point) -> bool:
         '''check if two poits are equal'''
         if not isinstance(pos, Point):
-            pos = Point._make(pos)
+            try:
+                pos = Point._make(pos)
+            except TypeError:
+                return NotImplemented
         return self.x == pos.x and self.y == pos.y
 
+    def __abs__(self) -> Point:
+        '''Returns the a Point with absolute value on x and y'''
+        return Point(abs(self.x), abs(self.y))
+
     @staticmethod
     def distance(pos1: Point, pos2: Point) -> float:
-        """takes two points and returns the distance between then"""
+        '''takes two points and returns the distance between then'''
         if not isinstance(pos1, Point):
             pos1 = Point._make(pos1)
         if not isinstance(pos2, Point):
@@ -178,7 +231,7 @@ class Point(RecordClass):
         return abs((end.x - start.x) * (start.y - point.y) - (start.x - point.x) * (end.y - start.y)) / math.sqrt((end.x - start.x) ** 2 + (end.y - start.y) ** 2)
 
 def clip_surface(surface: pygame.Surface, rect: Rect) -> pygame.Surface:
-    """Copy part of a pygame.Surface"""
+    '''Copy part of a pygame.Surface'''
     cropped = pygame.Surface(rect.size)
     cropped.blit(surface, (0, 0), rect)
     return cropped
@@ -208,7 +261,7 @@ def get_bezier_curve_points(p0: Point, p1: Point, p2: Point, density: int) -> Li
     return result
 
 class Animation:
-    """
+    '''
     Represents a object that has multiple frames each with diffrent length
     :example:
 
@@ -228,9 +281,9 @@ class Animation:
                 a.update()
 
         Example().run()
-    """
+    '''
     def __init__(self, glob_path: str, frame_data: [int], repititions: int = None):
-        """
+        '''
         :glob_path: the path that glob is called on.
             e.g.: 'assets/animations/*' to get every file in assets/animations
         :frame_data: how long a frame of the animation should be displayed in game frames
@@ -238,7 +291,7 @@ class Animation:
             this must be the same length as the number of items from glob_path
         :repititions: Optional. defaults to None. if repititions is none, it repeats forever.
             if this number is an int, it decrements every time update is called until it is zero
-        """
+        '''
         self.glob_path = glob_path
         self.frame_data = frame_data
         self.repititions = repititions
@@ -246,9 +299,9 @@ class Animation:
         self.load(glob_path, frame_data)
 
     def update(self):
-        """
+        '''
         Indicate a frame has passed
-        """
+        '''
         if not self.finished:
             self.frames_until_next -= 1
             if self.frames_until_next == 0:
@@ -260,23 +313,23 @@ class Animation:
                         self.finished = True
 
     def get_surface(self) -> pygame.Surface:
-        """return the frame of the current index"""
+        '''return the frame of the current index'''
         return self.frames[self.frame_index][0]
 
     def reset(self):
-        """Restart the animation to the start of the loop"""
+        '''Restart the animation to the start of the loop'''
         self.frame_index = 0
         self.frames_until_next = self.frames[0][1]
 
     def load(self, glob_path: str, frame_data):
-        """
+        '''
         Load animations from a glob path
         :glob_path: the path that glob is called on.
             e.g.: 'assets/animations/*' to get every file in assets/animations
         :frame_data: how long a frame of the animation should be displayed in game frames
             e.g.: [7, 8, 9] first image found in glob_path lasts 7, the next lasts 8, and the third lasts 9
             this must be the same length as the number of items from glob_path
-        """
+        '''
         file_names = glob(glob_path)
         if len(file_names) != len(frame_data):
             raise ValueError('Length of frame_data and the number of files must be the same')
@@ -353,7 +406,7 @@ class Particle(Circle):
                     self.alive = False
 
 class Button:
-    """A button in a pygame application"""
+    '''A button in a pygame application'''
 
     def __init__(
             self,
@@ -395,13 +448,13 @@ class Button:
         screen.blit(text_obj, (self.rect.centerx - text_size[0] / 2, self.rect.centery - text_size[1] / 2))
 
     def __call__(self):
-        """Overwrite the () operator on the button object"""
+        '''Overwrite the () operator on the button object'''
         if self.action:
             self.action()
         self.clicked = True
 
 class ToggleButton:
-    """When clickd this button will change its color, text, and also call target"""
+    '''When clickd this button will change its color, text, and also call target'''
 
     def __init__(
             self,
@@ -459,13 +512,13 @@ class ToggleButton:
             screen.blit(text_obj, (self.rect.centerx - text_size[0] / 2, self.rect.centery - text_size[1] / 2))
 
     def __call__(self):
-        """override the ()"""
+        '''override the ()'''
         if self.action:
             self.action()
         self.toggled = not self.toggled
 
 class GameScreen:
-    """
+    '''
     A class to reperesent a screen inside a pygame application
     e.g.: menu, pause screen, or main screen
     to use this class, inherit it and overwrite some/all of its functions
@@ -483,17 +536,17 @@ class GameScreen:
 
         example = Example()
         example.run()
-    """
+    '''
 
     def __init__(self, screen: pygame.Surface, real_window_size: Point, window_size: Point = None, frame_rate: int = 30):
-        """
+        '''
         :screen: The pygame surface that will be drawn onto
         :real_window_size: The height and width of the screen in real computer pixels
         :window_size: The height and width of the screen in game pixels pixels
             if this is smaller than real_window_size the pixels become larger
             if this is larger than real_window_size the pixels become smaller
         :frame_rate: The desired frame rate of the current screen
-        """
+        '''
         self.window_scaled = bool(window_size) and window_size != real_window_size
         self.real_screen = screen
         self.screen = screen if not self.window_scaled else pygame.Surface(window_size)
@@ -517,19 +570,19 @@ class GameScreen:
             self.game_ticks = 0
 
     def key_down(self, event: pygame.event.Event):
-        """Function called when a pygame KEYDOWN event is triggered"""
+        '''Function called when a pygame KEYDOWN event is triggered'''
 
     def key_up(self, event: pygame.event.Event):
-        """Function called when a pygame KEYUP event is triggered"""
+        '''Function called when a pygame KEYUP event is triggered'''
 
     def mouse_button_down(self, event: pygame.event.Event):
-        """Function called when a pygame MOUSEBUTTONDOWN event is triggered"""
+        '''Function called when a pygame MOUSEBUTTONDOWN event is triggered'''
 
     def mouse_button_up(self, event: pygame.event.Event):
-        """Function called when a pygame key_down MOUSEBUTTONDOWN is triggered"""
+        '''Function called when a pygame key_down MOUSEBUTTONDOWN is triggered'''
 
     def handle_event(self, event: pygame.event.Event):
-        """Handle a pygame events"""
+        '''Handle a pygame events'''
         if event.type == QUIT:
             sys.exit()
         elif event.type == KEYDOWN:
@@ -542,11 +595,11 @@ class GameScreen:
             self.mouse_button_up(event)
 
     def update(self):
-        """Run every frame, meant for drawing and update logic"""
+        '''Run every frame, meant for drawing and update logic'''
         self.screen.fill((0, 0, 100))
 
     def run(self):
-        """Run the main loop"""
+        '''Run the main loop'''
         self.running = True
         while self.running:
             for event in pygame.event.get():
@@ -558,10 +611,10 @@ class GameScreen:
             self.tick()
 
 class MenuScreen(GameScreen):
-    """
+    '''
     A class to represent a menu screen inside a pygame application
     e.g.: Main menu, Pause menu, Options
-    """
+    '''
 
     def __init__(self, screen: pygame.Surface, real_window_size: Point, window_size: Point = None, frame_rate: int = 30):
         super().__init__(screen, real_window_size, window_size, frame_rate)
@@ -583,7 +636,7 @@ class MenuScreen(GameScreen):
             self.buttons[self.button_index]()
 
     def draw_buttons(self, screen: pygame.Surface = None, highlight: bool = True):
-        """Draw the buttons"""
+        '''Draw the buttons'''
         if not screen:
             screen = self.screen
         for i, button in enumerate(self.buttons):
